"""
Cerberror
=========

The package converts errors generated by Cerberus
validator to customized messages defined by a user.

"""

__all__ = ["ErrConverter", "PathFinder", "Translator"]
__version__ = "0.1.1"
__author__ = "Przemysław Bruś"

from cerberror.errors import ErrConverter
from cerberror.paths import PathFinder
from cerberror.trans import Translator
